import importlib.util
import inspect
import logging
import os
import xmlschema
from xmlschema.validators.exceptions import XMLSchemaValidationError
from .operations import OperationAbstract

logger = logging.getLogger("fmu_manipulation_toolbox")


class OperationGenericCheck(OperationAbstract):
    SUPPORTED_FMI_VERSIONS = ('2.0', '3.0')

    def __init__(self):
        self.compliant_with_version = None

    def __repr__(self):
        return f"FMU Generic Conformity Checks"

    def fmi_attrs(self, attrs):
        if attrs['fmiVersion'] not in self.SUPPORTED_FMI_VERSIONS:
            logger.error(f"Expected FMI {','.join(self.SUPPORTED_FMI_VERSIONS)} versions.")
            return

        fmi_name = f"fmi{attrs['fmiVersion'][0]}"

        xsd_filename = os.path.join(os.path.dirname(__file__), "resources", "fmi-" + attrs['fmiVersion'],
                                    f"{fmi_name}ModelDescription.xsd")
        try:
            xmlschema.validate(self.fmu.descriptor_filename, schema=xsd_filename)
        except XMLSchemaValidationError as error:
            logger.error(error.reason, error.msg)
        else:
            self.compliant_with_version = attrs['fmiVersion']

    def closure(self):
        if self.compliant_with_version:
            logger.info(f"This FMU seems to be compliant with FMI-{self.compliant_with_version}.")
        else:
            logger.error(f"This FMU does not validate with FMI standard.")


checker_list = [OperationGenericCheck]


def add_from_file(checker_filename: str):
    spec = importlib.util.spec_from_file_location(checker_filename, checker_filename)
    if not spec:
        logger.error(f"Cannot load '{checker_filename}'. Is this a python file?")
        return
    try:
        checker_module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(checker_module)
        except (ModuleNotFoundError, SyntaxError) as error:
            logger.error(f"Cannot load '{checker_filename}': {error})")
            return

        for checker_name, checker_class in inspect.getmembers(checker_module, inspect.isclass):
            if OperationAbstract in checker_class.__bases__:
                checker_list.append(checker_class)
                logger.info(f"Adding checker: {checker_filename}|{checker_name}")

    except AttributeError:
        logger.error(f"'{checker_filename}' should implement class 'OperationCheck'")



import csv
import html
import logging
import os
import re
import shutil
import tempfile
import xml.parsers.expat
import zipfile
import hashlib
from typing import *

logger = logging.getLogger("fmu_manipulation_toolbox")

class FMU:
    """Unpack and Repack facilities for FMU package. Once unpacked, we can process Operation on
    modelDescription.xml file."""

    FMI2_TYPES = ('Real', 'Integer', 'String', 'Boolean', 'Enumeration')
    FMI3_TYPES = ('Float64', 'Float32',
                  'Int8', 'UInt8', 'Int16', 'UInt16', 'Int32', 'UInt32', 'Int64', 'UInt64',
                  'String', 'Boolean', 'Enumeration')

    def __init__(self, fmu_filename):
        self.fmu_filename = fmu_filename
        self.tmp_directory = tempfile.mkdtemp()
        self.fmi_version = None

        try:
            with zipfile.ZipFile(self.fmu_filename) as zin:
                zin.extractall(self.tmp_directory)
        except FileNotFoundError:
            raise FMUError(f"'{fmu_filename}' does not exist")
        self.descriptor_filename = os.path.join(self.tmp_directory, "modelDescription.xml")
        if not os.path.isfile(self.descriptor_filename):
            raise FMUError(f"'{fmu_filename}' is not valid: {self.descriptor_filename} not found")

    def __del__(self):
        shutil.rmtree(self.tmp_directory)

    def save_descriptor(self, filename):
        shutil.copyfile(os.path.join(self.tmp_directory, "modelDescription.xml"), filename)

    def repack(self, filename):
        with zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED) as zout:
            for root, dirs, files in os.walk(self.tmp_directory):
                for file in files:
                    zout.write(os.path.join(root, file),
                               os.path.relpath(os.path.join(root, file), self.tmp_directory))
        # TODO: Add check on output file

    def apply_operation(self, operation, apply_on=None):
        manipulation = Manipulation(operation, self)
        manipulation.manipulate(self.descriptor_filename, apply_on)


class FMUPort:
    def __init__(self):
        self.fmi_type = None
        self.attrs_list: List[Dict] = []
        self.dimension = None

    def dict_level(self, nb):
        return " ".join([f'{key}="{value}"' for key, value in self.attrs_list[nb].items()])

    def write_xml(self, fmi_version: int, file):
        if fmi_version == 2:
            print(f"    <ScalarVariable {self.dict_level(0)}>", file=file)
            print(f"      <{self.fmi_type} {self.dict_level(1)}/>", file=file)
            print(f"    </ScalarVariable>", file=file)
        elif fmi_version == 3:
            print(f"    <{self.fmi_type} {self.dict_level(0)}/>", file=file)
        else:
            raise FMUError(f"FMUPort writing: unsupported FMI version {fmi_version}")

    def __contains__(self, item):
        for attrs in self.attrs_list:
            if item in attrs:
                return True
        return False

    def __getitem__(self, item):
        for attrs in self.attrs_list:
            if item in attrs:
                return attrs[item]
        raise KeyError

    def __setitem__(self, key, value):
        for attrs in self.attrs_list:
            if key in attrs:
                attrs[key] = value
                return
        raise KeyError

    def get(self, item, default_value):
        try:
            return self[item]
        except KeyError:
            return default_value

    def push_attrs(self, attrs):
        self.attrs_list.append(attrs)


class FMUError(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __repr__(self):
        return self.reason


class Manipulation:
    """Parse modelDescription.xml file and create a modified version"""
    def __init__(self, operation, fmu):
        self.output_filename = tempfile.mktemp()
        self.out = None
        self.operation = operation
        self.parser = xml.parsers.expat.ParserCreate()
        self.parser.StartElementHandler = self.start_element
        self.parser.EndElementHandler = self.end_element
        self.parser.CharacterDataHandler = self.char_data
        self.skip_until: Optional[str] = None
        self.operation.set_fmu(fmu)
        self.fmu = fmu

        self.current_port: Optional[FMUPort] = None

        self.current_port_number: int = 0
        self.port_translation: List[Optional[int]] = []
        self.port_names_list: List[str] = []
        self.port_removed_vr: Set[str] = set()
        self.apply_on = None

    @staticmethod
    def escape(value):
        if isinstance(value, str):
            return html.escape(html.unescape(value))
        else:
            return value

    def handle_port(self):
        causality = self.current_port.get('causality', 'local')
        port_name = self.current_port['name']
        vr = self.current_port['valueReference']
        if not self.apply_on or causality in self.apply_on:
            if self.operation.port_attrs(self.current_port):
                self.remove_port(port_name, vr)
                # Exception is raised by remove port !
            else:
                self.keep_port(port_name)
        else:  # Keep ScalarVariable as it is.
            self.keep_port(port_name)

    def start_element(self, name, attrs):
        if self.skip_until:
            return
        try:
            if name == 'ScalarVariable': # FMI 2.0 only
                self.current_port = FMUPort()
                self.current_port.push_attrs(attrs)
            elif self.fmu.fmi_version == 2 and name in self.fmu.FMI2_TYPES:
                if self.current_port: # <Enumeration> can be found before port defition. Ignored.
                    self.current_port.fmi_type = name
                    self.current_port.push_attrs(attrs)
            elif self.fmu.fmi_version == 3 and name in self.fmu.FMI3_TYPES:
                self.current_port = FMUPort()
                self.current_port.fmi_type = name
                self.current_port.push_attrs(attrs)
            elif name == 'CoSimulation':
                self.operation.cosimulation_attrs(attrs)
            elif name == 'DefaultExperiment':
                self.operation.experiment_attrs(attrs)
            elif name == 'fmiModelDescription':
                self.fmu.fmi_version = int(float(attrs["fmiVersion"]))
                self.operation.fmi_attrs(attrs)
            elif name == 'Unknown': # FMI-2.0 only
                self.unknown_attrs(attrs)
            elif name == 'Output' or name == "ContinuousStateDerivative" or "InitialUnknown":
                self.handle_structure(attrs)

        except ManipulationSkipTag:
            self.skip_until = name
            return

        if self.current_port is None:
            if attrs:
                attrs_list = [f'{key}="{self.escape(value)}"' for (key, value) in attrs.items()]
                print(f"<{name}", " ".join(attrs_list), ">", end='', file=self.out)
            else:
                print(f"<{name}>", end='', file=self.out)

    def end_element(self, name):
        if self.skip_until:
            if self.skip_until == name:
                self.skip_until = None
            return
        else:
            if name == "ScalarVariable" or (self.fmu.fmi_version == 3 and name in FMU.FMI3_TYPES):
                try:
                    self.handle_port()
                    self.current_port.write_xml(self.fmu.fmi_version, self.out)
                except ManipulationSkipTag:
                    logger.info(f"Port '{self.current_port['name']}' is removed.")
                self.current_port = None

            elif self.current_port is None:
                print(f"</{name}>", end='', file=self.out)

    def char_data(self, data):
        if not self.skip_until:
            print(data, end='', file=self.out)

    def remove_port(self, name, vr):
        self.port_names_list.append(name)
        self.port_translation.append(None)
        self.port_removed_vr.add(vr)
        raise ManipulationSkipTag

    def keep_port(self, name):
        self.port_names_list.append(name)
        self.current_port_number += 1
        self.port_translation.append(self.current_port_number)

    def unknown_attrs(self, attrs):
        index = int(attrs['index']) - 1
        new_index = self.port_translation[index]
        if new_index:
            attrs['index'] = self.port_translation[int(attrs['index']) - 1]
        else:
            logger.warning(f"Removed port '{self.port_names_list[index]}' is involved in dependencies tree.")
            raise ManipulationSkipTag

    def handle_structure(self, attrs):
        try:
            vr = attrs['valueReference']
            if vr in self.port_removed_vr:
                raise ManipulationSkipTag
        except KeyError:
            return

        try:
            dependencies = attrs['dependencies']
            new_dependencies = []
            for dependency in dependencies.split(" "):
                if dependency in self.port_removed_vr:
                    logger.warning(f"Removed port 'vr={dependency}' was involved in dependencies tree.")
                else:
                    new_dependencies.append(dependency)
            attrs['dependencies'] = " ".join(new_dependencies)
        except KeyError:
            return

    def manipulate(self, descriptor_filename, apply_on=None):
        self.apply_on = apply_on
        with open(self.output_filename, "w", encoding="utf-8") as self.out, open(descriptor_filename, "rb") as file:
            self.parser.ParseFile(file)
        self.operation.closure()
        os.replace(self.output_filename, descriptor_filename)


class ManipulationSkipTag(Exception):
    """Exception: We need to skip every thing until matching closing tag"""


class OperationAbstract:
    """This class hold hooks called during parsing"""
    fmu: FMU = None

    def set_fmu(self, fmu):
        self.fmu = fmu

    def fmi_attrs(self, attrs):
        pass

    def cosimulation_attrs(self, attrs):
        pass

    def experiment_attrs(self, attrs):
        pass

    def port_attrs(self, fmu_port: FMUPort) -> int:
        """ return 0 to keep port, otherwise remove it"""
        return 0

    def closure(self):
        pass


class OperationSaveNamesToCSV(OperationAbstract):
    def __repr__(self):
        return f"Dump names into '{self.output_filename}'"

    def __init__(self, filename):
        self.output_filename = filename
        self.csvfile = open(filename, 'w', newline='')
        self.writer = csv.writer(self.csvfile, delimiter=';', quotechar="'", quoting=csv.QUOTE_MINIMAL)
        self.writer.writerow(['name', 'newName', 'valueReference', 'causality', 'variability', 'scalarType',
                              'startValue'])

    def closure(self):
        self.csvfile.close()

    def port_attrs(self, fmu_port: FMUPort) -> int:
        self.writer.writerow([fmu_port["name"],
                              fmu_port["name"],
                              fmu_port["valueReference"],
                              fmu_port.get("causality", "local"),
                              fmu_port.get("variability", "continuous"),
                              fmu_port.fmi_type,
                              fmu_port.get("start", "")])

        return 0


class OperationStripTopLevel(OperationAbstract):
    def __repr__(self):
        return "Remove Top Level Bus"

    def port_attrs(self, fmu_port):
        new_name = fmu_port['name'].split('.', 1)[-1]
        fmu_port['name'] = new_name
        return 0


class OperationMergeTopLevel(OperationAbstract):
    def __repr__(self):
        return "Merge Top Level Bus with signal names"

    def port_attrs(self, fmu_port):
        old = fmu_port['name']
        fmu_port['name'] = old.replace('.', '_', 1)
        return 0


class OperationRenameFromCSV(OperationAbstract):
    def __repr__(self):
        return f"Rename according to '{self.csv_filename}'"

    def __init__(self, csv_filename):
        self.csv_filename = csv_filename
        self.translations = {}

        try:
            with open(csv_filename, newline='') as csvfile:
                reader = csv.reader(csvfile, delimiter=';', quotechar="'")
                for row in reader:
                    self.translations[row[0]] = row[1]
        except FileNotFoundError:
            raise OperationError(f"file '{csv_filename}' is not found")
        except KeyError:
            raise OperationError(f"file '{csv_filename}' should contain two columns")

    def port_attrs(self, fmu_port):
        name = fmu_port['name']
        try:
            new_name = self.translations[fmu_port['name']]
        except KeyError:
            new_name = name  # if port is not in CSV file, keep old name

        if new_name:
            fmu_port['name'] = new_name
            return 0
        else:
            # we want to delete this name!
            return 1


class OperationRemoveRegexp(OperationAbstract):
    def __repr__(self):
        return f"Remove ports matching '{self.regex_string}'"

    def __init__(self, regex_string):
        self.regex_string = regex_string
        self.regex = re.compile(regex_string)
        self.current_port_number = 0
        self.port_translation = []

    def port_attrs(self, fmu_port):
        name = fmu_port['name']
        if self.regex.match(name):
            return 1  # Remove port
        else:
            return 0


class OperationKeepOnlyRegexp(OperationAbstract):
    def __repr__(self):
        return f"Keep only ports matching '{self.regex_string}'"

    def __init__(self, regex_string):
        self.regex_string = regex_string
        self.regex = re.compile(regex_string)

    def port_attrs(self, fmu_port):
        name = fmu_port['name']
        if self.regex.match(name):
            return 0
        else:
            return 1  # Remove port


class OperationSummary(OperationAbstract):
    def __init__(self):
        self.nb_port_per_causality = {}

    def __repr__(self):
        return f"FMU Summary"

    def fmi_attrs(self, attrs):
        logger.info(f"| fmu filename = {self.fmu.fmu_filename}")
        logger.info(f"| temporary directory = {self.fmu.tmp_directory}")
        hash_md5 = hashlib.md5()
        with open(self.fmu.fmu_filename, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        digest = hash_md5.hexdigest()
        logger.info(f"| MD5Sum = {digest}")

        logger.info(f"|\n| FMI properties: ")
        for (k, v) in attrs.items():
            logger.info(f"|  - {k} = {v}")
        logger.info(f"|")

    def cosimulation_attrs(self, attrs):
        logger.info("| Co-Simulation capabilities: ")
        for (k, v) in attrs.items():
            logger.info(f"|  - {k} = {v}")
        logger.info(f"|")

    def experiment_attrs(self, attrs):
        logger.info("| Default Experiment values: ")
        for (k, v) in attrs.items():
            logger.info(f"|  - {k} = {v}")
        logger.info(f"|")

    def port_attrs(self, fmu_port) -> int:
        causality = fmu_port.get("causality", "local")

        try:
            self.nb_port_per_causality[causality] += 1
        except KeyError:
            self.nb_port_per_causality[causality] = 1

        return 0

    def closure(self):
        logger.info("| Supported platforms: ")
        try:
            for platform in os.listdir(os.path.join(self.fmu.tmp_directory, "binaries")):
                logger.info(f"|  - {platform}")
        except FileNotFoundError:
            pass  # no binaries

        if os.path.isdir(os.path.join(self.fmu.tmp_directory, "sources")):
            logger.info(f"|  - RT (sources available)")

        resource_dir = os.path.join(self.fmu.tmp_directory, "resources")
        if os.path.isdir(resource_dir):
            logger.info("|\n| Embedded resources:")
            for resource in os.listdir(resource_dir):
                logger.info(f"|  - {resource}")

        extra_dir = os.path.join(self.fmu.tmp_directory, "extra")
        if os.path.isdir(extra_dir):
            logger.info("|\n| Additional (meta-)data:")
            for extra in os.listdir(extra_dir):
                logger.info(f"|  - {extra}")

        logger.info("|\n| Number of ports")
        for causality, nb_ports in self.nb_port_per_causality.items():
            logger.info(f"|  {causality} : {nb_ports}")

        logger.info("|\n| [End of report]")


class OperationRemoveSources(OperationAbstract):
    def __repr__(self):
        return f"Remove sources"

    def cosimulation_attrs(self, attrs):
        try:
            shutil.rmtree(os.path.join(self.fmu.tmp_directory, "sources"))
        except FileNotFoundError:
            logger.info("This FMU does not embed sources.")


class OperationTrimUntil(OperationAbstract):
    def __init__(self, separator):
        self.separator = separator

    def __repr__(self):
        return f"Trim names until (and including) '{self.separator}'"

    def port_attrs(self, fmu_port) -> int:
        name = fmu_port['name']
        try:
            fmu_port['name'] = name[name.index(self.separator)+len(self.separator):-1]
        except KeyError:
            pass  # no separator

        return 0


class OperationError(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __repr__(self):
        return self.reason



import argparse
import logging

from .utils import setup_logger, make_wide
from ..split import FMUSplitter, FMUSplitterError
from ..version import __version__ as version


def fmusplit():
    logger = setup_logger()

    logger.info(f"FMUSplit version {version}")
    parser = argparse.ArgumentParser(prog="fmusplit", description="Split FMU Container into FMU's",
                                     formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
                                     add_help=False,
                                     epilog="see: https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/"
                                            "container/README.md")
    parser.add_argument('-h', '-help', action="help")

    parser.add_argument("-debug", action="store_true", dest="debug",
                        help="Add lot of useful log during the process.")

    parser.add_argument("-fmu", action="append", dest="fmu_filename_list", default=[],
                        metavar="filename.fmu", required=True,
                        help="Description of the FMU container to split.")

    config = parser.parse_args()

    if config.debug:
        logger.setLevel(logging.DEBUG)

    for fmu_filename in config.fmu_filename_list:
        try:
            splitter = FMUSplitter(fmu_filename)
            splitter.split_fmu()
        except FMUSplitterError as e:
            logger.fatal(f"{fmu_filename}: {e}")
            continue
        except FileNotFoundError as e:
            logger.fatal(f"Cannot read file: {e}")
            continue


if __name__ == "__main__":
    fmusplit()



import argparse
import logging

from pathlib import Path

from .utils import setup_logger, make_wide
from ..assembly import Assembly, AssemblyError
from ..container import FMUContainerError
from ..version import __version__ as version


def fmucontainer():
    logger = setup_logger()

    logger.info(f"FMUContainer version {version}")
    parser = argparse.ArgumentParser(prog="fmucontainer", description="Generate FMU from FMU's",
                                     formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
                                     add_help=False,
                                     epilog="see: https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/"
                                            "container/README.md")

    parser.add_argument('-h', '-help', action="help")

    parser.add_argument("-fmu-directory", action="store", dest="fmu_directory", required=False, default=".",
                        help="Directory containing initial FMU’s and used to generate containers. "
                             "If not defined, current directory is used.")

    parser.add_argument("-fmi", action="store", dest="fmi_version", required=False, default="2",
                        help="Define version of FMI to be used for container interface."
                             "Only '2' or '3' is supported.")

    parser.add_argument("-container", action="append", dest="container_descriptions_list", default=[],
                        metavar="filename.{csv|json|ssp},[:step_size]", required=True,
                        help="Description of the container to create.")

    parser.add_argument("-debug", action="store_true", dest="debug",
                        help="Add lot of useful log during the process.")

    parser.add_argument("-no-auto-input", action="store_false", dest="auto_input", default=True,
                        help="Create ONLY explicit input.")

    parser.add_argument("-no-auto-output", action="store_false", dest="auto_output", default=True,
                        help="Create ONLY explicit output.")

    parser.add_argument("-auto-parameter", action="store_true", dest="auto_parameter", default=False,
                        help="Expose parameters of the embedded fmu's.")

    parser.add_argument("-auto-local", action="store_true", dest="auto_local", default=False,
                        help="Expose local variables of the embedded fmu's.")

    parser.add_argument("-no-auto-link", action="store_false", dest="auto_link", default=True,
                        help="Create ONLY explicit links.")

    parser.add_argument("-mt", action="store_true", dest="mt", default=False,
                        help="Enable Multi-Threaded mode for the generated container.")

    parser.add_argument("-profile", action="store_true", dest="profiling", default=False,
                        help="Enable Profiling mode for the generated container.")

    parser.add_argument("-sequential", action="store_true", dest="sequential", default=False,
                        help="Use sequential mode to schedule embedded fmu's.")

    parser.add_argument("-dump-json",  action="store_true", dest="dump", default=False,
                        help="Dump a JSON file for each container.")

    config = parser.parse_args()

    if config.debug:
        logger.setLevel(logging.DEBUG)

    fmu_directory = Path(config.fmu_directory)
    logger.info(f"FMU directory: '{fmu_directory}'")

    for description in config.container_descriptions_list:
        try:
            tokens = description.split(":")
            filename = ":".join(tokens[:-1])
            step_size = float(tokens[-1])
        except ValueError:
            step_size = None
            filename = description

        try:
            assembly = Assembly(filename, step_size=step_size, auto_link=config.auto_link,
                                auto_input=config.auto_input, auto_output=config.auto_output,
                                auto_local=config.auto_local, mt=config.mt, sequential=config.sequential,
                                profiling=config.profiling, fmu_directory=fmu_directory, debug=config.debug,
                                auto_parameter=config.auto_parameter)
        except FileNotFoundError as e:
            logger.fatal(f"Cannot read file: {e}")
            continue
        except (FMUContainerError, AssemblyError) as e:
            logger.fatal(f"{filename}: {e}")
            continue

        try:
            assembly.make_fmu(dump_json=config.dump, fmi_version=int(config.fmi_version))
        except FMUContainerError as e:
            logger.fatal(f"{filename}: {e}")
            continue

if __name__ == "__main__":
    fmucontainer()



import logging
import sys
from colorama import Fore, Style, init

def setup_logger():
    class CustomFormatter(logging.Formatter):
        def format(self, record):
            log_format = "%(levelname)-8s | %(message)s"
            format_per_level = {
                logging.DEBUG: str(Fore.BLUE) + log_format,
                logging.INFO: str(Fore.CYAN) + log_format,
                logging.WARNING: str(Fore.YELLOW) + log_format,
                logging.ERROR: str(Fore.RED) + log_format,
                logging.CRITICAL: str(Fore.RED + Style.BRIGHT) + log_format,
            }
            formatter = logging.Formatter(format_per_level[record.levelno])
            return formatter.format(record)
    init()
    logger = logging.getLogger("fmu_manipulation_toolbox")
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(CustomFormatter())
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    return logger

def make_wide(formatter, w=120, h=36):
    """Return a wider HelpFormatter, if possible."""
    try:
        # https://stackoverflow.com/a/5464440
        # beware: "Only the name of this class is considered a public API."
        kwargs = {'width': w, 'max_help_position': h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        return formatter



import os

if os.name == 'nt':
    gui_style = """
    QWidget {
        font: 10pt "Verdana";
        background: #4b4e51;
        color: #b5bab9;
    }
    QPushButton, QComboBox {
        min-height: 30px;
        padding: 1px 1px 0.2em 0.2em;
        border: 1px solid #282830;
        border-radius: 5px;
        color: #dddddd;
    }
    QPushButton:pressed {
        border: 2px solid #282830;
    }
    QPushButton.info {
        background-color: #4e6749;
    }
    QPushButton.info:hover {
        background-color: #5f7850;
    }
    QPushButton.modify {
        background-color: #98763f;
    }
    QPushButton.modify:hover {
        background-color: #a9874f;
    }
    QPushButton.removal {
        background-color: #692e2e;
    }
    QPushButton.removal:hover {
        background-color: #7a3f3f;
    }
    QPushButton.save {
        background-color: #564967;
    }
    QPushButton.save:hover {
        background-color: #675a78;
    }
    QPushButton.quit {
        background-color: #4571a4;
    }
    QPushButton.quit:hover {
        background-color: #5682b5;
    }
    QPushButton::disabled {
        background-color: gray;
    }
    QToolTip {
        color: black
    }
    QLabel.dropped_fmu {
        background-color: #b5bab9
    }
    QLabel.title {
        font: 14pt bold "Verdana";
    }
    QLabel.dropped_fmu:hover {
        background-color: #c6cbca
    }
    QTextBrowser, QTreeView {
        font: 11pt "Consolas";
        background-color: #282830;
        color: #b5bab9;
    }
    QMenu::item {
        padding: 2px 250px 2px 20px;
        border: 1px solid transparent;
    }
    QMenu::item::indicator, QCheckBox::item::indicator {
        width: 32px;
        height: 32px;
    }
    QMenu::indicator:checked, QCheckBox::indicator:checked {
        image: url(images:checkbox-checked.png);
    }
    QMenu::indicator:checked:hover, QCheckBox::indicator:checked:hover {
        image: url(images:checkbox-checked-hover.png);
    }
    QMenu::indicator:checked:disabled, QCheckBox::indicator:checked:disabled {
        image: url(images:checkbox-checked-disabled.png);
    }
    QMenu::indicator:unchecked, QCheckBox::indicator:unchecked {
        image: url(images:checkbox-unchecked.png);
    }
    QMenu::indicator:unchecked:hover, QCheckBox::indicator:unchecked:hover {
        image: url(images:checkbox-unchecked-hover.png);
    }
    QMenu::indicator:unchecked:disabled, QCheckBox::indicator:unchecked:disabled {
        image: url(images:checkbox-unchecked-disabled.png);
    }
    QCheckBox::item {
        padding: 2px 250px 2px 20px;
        border: 1px solid transparent;
    }
    QTabBar::tab {
        min-height: 30px;
        padding: 1px 1px 0.2em 0.2em;
        color: #dddddd;
        margin: 2px;
        margin-bottom: 0px;
        border: 1px solid #282830;
        border-top-left-radius: 5px;
        border-top-right-radius: 5px;
    } 
    QTabBar::tab:selected, QTabBar::tab:hover {
        background-color: #5f7850;
        margin-bottom:-1px;
    }
    QTabBar {
        border-bottom: 1px solid #282830;
    }
    QTabBar::tab:top:last, QTabBar::tab:bottom:last {
        margin-right: 0;
    }
    QTabBar::tab:top:first, QTabBar::tab:bottom:first {
        margin-left: 0;
    }
    """
else:
    gui_style = """
    QWidget {
        font: 12pt;
        background: #4b4e51;
        color: #b5bab9;
    }
    QPushButton, QComboBox {
        min-height: 30px;
        padding: 1px 1px 0.2em 0.2em;
        border: 1px solid #282830;
        border-radius: 5px;
        color: #dddddd;
    }
    QPushButton:pressed {
        border: 2px solid #282830;
    }
    QPushButton.info {
        background-color: #4e6749;
    }
    QPushButton.info:hover {
        background-color: #5f7850;
    }
    QPushButton.modify {
        background-color: #98763f;
    }
    QPushButton.modify:hover {
        background-color: #a9874f;
    }
    QPushButton.removal {
        background-color: #692e2e;
    }
    QPushButton.removal:hover {
        background-color: #7a3f3f;
    }
    QPushButton.save {
        background-color: #564967;
    }
    QPushButton.save:hover {
        background-color: #675a78;
    }
    QPushButton.quit {
        background-color: #4571a4;
    }
    QPushButton.quit:hover {
        background-color: #5682b5;
    }
    QPushButton::disabled {
        background-color: gray;
    }
    QToolTip {
        color: black
    }
    QLabel.dropped_fmu {
        background-color: #b5bab9
    }
    QLabel.title {
        font: 14pt bold "Verdana";
    }
    QLabel.dropped_fmu:hover {
        background-color: #c6cbca
    }
    QTextBrowser, QTreeView {
        font: 14pt "Courier New";
        background-color: #282830;
        color: #b5bab9;
    }
    QMenu::item {
        padding: 2px 250px 2px 20px;
        border: 1px solid transparent;
    }
    QMenu::item::indicator, QCheckBox::item::indicator {
        width: 32px;
        height: 32px;
    }
    QMenu::indicator:checked, QCheckBox::indicator:checked {
        image: url(images:checkbox-checked.png);
    }
    QMenu::indicator:checked:hover, QCheckBox::indicator:checked:hover {
        image: url(images:checkbox-checked-hover.png);
    }
    QMenu::indicator:checked:disabled, QCheckBox::indicator:checked:disabled {
        image: url(images:checkbox-checked-disabled.png);
    }
    QMenu::indicator:unchecked, QCheckBox::indicator:unchecked {
        image: url(images:checkbox-unchecked.png);
    }
    QMenu::indicator:unchecked:hover, QCheckBox::indicator:unchecked:hover {
        image: url(images:checkbox-unchecked-hover.png);
    }
    QMenu::indicator:unchecked:disabled, QCheckBox::indicator:unchecked:disabled {
        image: url(images:checkbox-unchecked-disabled.png);
    }
    QCheckBox::item {
        padding: 2px 250px 2px 20px;
        border: 1px solid transparent;
    }
    QTabBar::tab {
        min-height: 30px;
        padding: 1px 1px 0.2em 0.2em;
        color: #dddddd;
        margin: 2px;
        margin-bottom: 0px;
        border: 1px solid #282830;
        border-top-left-radius: 5px;
        border-top-right-radius: 5px;
    } 
    QTabBar::tab:selected, QTabBar::tab:hover {
        background-color: #5f7850;
        margin-bottom:-1px;
    }
    QTabBar {
        border-bottom: 1px solid #282830;
    }
    QTabBar::tab:top:last, QTabBar::tab:bottom:last {
        margin-right: 0;
    }
    QTabBar::tab:top:first, QTabBar::tab:bottom:first {
        margin-left: 0;
    }
    """

log_color = {
    "DEBUG": "#6E6B6B",
    "INFO": "#b5bab9",
    "WARNING": "#F7C61B",
    "ERROR": "#F54927",
    "CRITICAL": "#FF00FF",
}



import logging
import getpass
import math
import os
import shutil
import uuid
import platform
import zipfile
from datetime import datetime
from pathlib import Path
from typing import *

from .operations import FMU, OperationAbstract, FMUError, FMUPort
from .version import __version__ as tool_version


logger = logging.getLogger("fmu_manipulation_toolbox")


class EmbeddedFMUPort:
    FMI_TO_CONTAINER = {
        2: {
            'Real': 'real64',
            'Integer': 'integer32',
            'String': 'string',
            'Boolean': 'boolean'
        },
        3: {
            'Float64': 'real64',
            'Float32': 'real32',
            'Int8': 'integer8',
            'UInt8': 'uinteger8',
            'Int16': 'integer16',
            'UInt16': 'uinteger16',
            'Int32': 'integer32',
            'UInt32': 'uinteger32',
            'Int64': 'integer64',
            'UInt64': 'uinteger64',
            'String': 'string',
            'Boolean': 'boolean1'
        }
    }

    CONTAINER_TO_FMI = {
        2: {
            'real64': 'Real',
            'integer32': 'Integer',
            'string': 'String',
            'boolean': 'Boolean'
        },
        3: {
            'real64': 'Float64' ,
            'real32': 'Float32' ,
            'integer8': 'Int8' ,
            'uinteger8': 'UInt8' ,
            'integer16': 'Int16' ,
            'uinteger16': 'UInt16' ,
            'integer32': 'Int32' ,
            'uinteger32': 'UInt32' ,
            'integer64': 'Int64' ,
            'uinteger64': 'UInt64' ,
            'string': 'String' ,
            'boolean1': 'Boolean'
        }
    }

    ALL_TYPES = (
        "real64", "real32",
        "integer8", "uinteger8", "integer16", "uinteger16", "integer32", "uinteger32", "integer64", "uinteger64",
        "boolean", "boolean1",
        "string"
    )

    def __init__(self, fmi_type, attrs: Union[FMUPort, Dict[str, str]], fmi_version=0):
        self.causality = attrs.get("causality", "local")
        self.variability = attrs.get("variability", "continuous")
        self.name = attrs["name"]
        self.vr = int(attrs["valueReference"])
        self.description = attrs.get("description", None)

        if fmi_version > 0:
            self.type_name = self.FMI_TO_CONTAINER[fmi_version][fmi_type]
        else:
            self.type_name = fmi_type

        self.start_value = attrs.get("start", None)
        self.initial = attrs.get("initial", None)

    def xml(self, vr: int, name=None, causality=None, start=None, fmi_version=2) -> str:
        if name is None:
            name = self.name
        if causality is None:
            causality = self.causality
        if start is None:
            start = self.start_value
        if self.variability is None:
            self.variability = "continuous" if "real" in self.type_name else "discrete"

        try:
            fmi_type = self.CONTAINER_TO_FMI[fmi_version][self.type_name]
        except KeyError:
            logger.error(f"Cannot expose '{name}' because type '{self.type_name}' is not compatible "
                         f"with FMI-{fmi_version}.0")
            return ""

        if fmi_version == 2:
            child_attrs =  {
                "start": start,
            }

            filtered_child_attrs = {key: value for key, value in child_attrs.items() if value is not None}
            child_str = (f"<{fmi_type} " +
                         " ".join([f'{key}="{value}"' for (key, value) in filtered_child_attrs.items()]) +
                         "/>")

            scalar_attrs = {
                "name": name,
                "valueReference": vr,
                "causality": causality,
                "variability": self.variability,
                "initial": self.initial,
                "description": self.description,
            }
            filtered_attrs = {key: value for key, value in scalar_attrs.items() if value is not None}
            scalar_attrs_str = " ".join([f'{key}="{value}"' for (key, value) in filtered_attrs.items()])
            return f'<ScalarVariable {scalar_attrs_str}>{child_str}</ScalarVariable>'
        else:
            if fmi_type in ('String', 'Binary'):
                if start:
                    child_str = f'<Start value="{start}"/>'
                else:
                    child_str = ''
                scalar_attrs = {
                    "name": name,
                    "valueReference": vr,
                    "causality": causality,
                    "variability": self.variability,
                    "initial": self.initial,
                    "description": self.description,
                }
                filtered_attrs = {key: value for key, value in scalar_attrs.items() if value is not None}
                scalar_attrs_str = " ".join([f'{key}="{value}"' for (key, value) in filtered_attrs.items()])
                return f'<{fmi_type} {scalar_attrs_str}>{child_str}</{fmi_type}>'
            else:
                scalar_attrs = {
                    "name": name,
                    "valueReference": vr,
                    "causality": causality,
                    "variability": self.variability,
                    "initial": self.initial,
                    "description": self.description,
                    "start": start
                }
                filtered_attrs = {key: value for key, value in scalar_attrs.items() if value is not None}
                scalar_attrs_str = " ".join([f'{key}="{value}"' for (key, value) in filtered_attrs.items()])

                return f'<{fmi_type} {scalar_attrs_str}/>'


class EmbeddedFMU(OperationAbstract):
    capability_list = ("needsExecutionTool",
                       "canBeInstantiatedOnlyOncePerProcess",
                       "canHandleVariableCommunicationStepSize")

    def __init__(self, filename):
        self.fmu = FMU(filename)
        self.name = Path(filename).name
        self.id = Path(filename).stem.lower()

        self.step_size = None
        self.start_time = None
        self.stop_time = None
        self.model_identifier = None
        self.guid = None
        self.fmi_version = None
        self.ports: Dict[str, EmbeddedFMUPort] = {}

        self.capabilities: Dict[str, str] = {}
        self.current_port = None  # used during apply_operation()

        self.fmu.apply_operation(self)  # Should be the last command in constructor!
        if self.model_identifier is None:
            raise FMUContainerError(f"FMU '{self.name}' does not implement Co-Simulation mode.")

    def fmi_attrs(self, attrs):
        fmi_version = attrs['fmiVersion']
        if fmi_version == "2.0":
            self.guid = attrs['guid']
            self.fmi_version = 2
        if fmi_version == "3.0": # TODO: handle 3.x cases
            self.guid = attrs['instantiationToken']
            self.fmi_version = 3

    def cosimulation_attrs(self, attrs: Dict[str, str]):
        self.model_identifier = attrs['modelIdentifier']
        for capability in self.capability_list:
            self.capabilities[capability] = attrs.get(capability, "false")

    def experiment_attrs(self, attrs: Dict[str, str]):
        try:
            self.step_size = float(attrs['stepSize'])
        except KeyError:
            logger.warning(f"FMU '{self.name}' does not specify preferred step size")
        self.start_time = float(attrs.get("startTime", 0.0))
        self.stop_time = float(attrs.get("stopTime", self.start_time + 1.0))

    def port_attrs(self, fmu_port: FMUPort):
        # Container will manage Enumeration as Integer
        if fmu_port.fmi_type == "Enumeration":
            if self.fmi_version == 2:
                fmu_port.fmi_type = "Integer"
            else:
                fmu_port.fmi_type = "Int32"
        port = EmbeddedFMUPort(fmu_port.fmi_type, fmu_port, fmi_version=self.fmi_version)
        self.ports[port.name] = port

    def __repr__(self):
        return f"FMU '{self.name}' ({len(self.ports)} variables, ts={self.step_size}s)"


class FMUContainerError(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __repr__(self):
        return f"{self.reason}"


class ContainerPort:
    def __init__(self, fmu: EmbeddedFMU, port_name: str):
        self.fmu = fmu
        try:
            self.port = fmu.ports[port_name]
        except KeyError:
            raise FMUContainerError(f"Port '{fmu.name}/{port_name}' does not exist")
        self.vr = None

    def __repr__(self):
        return f"Port {self.fmu.name}/{self.port.name}"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)


class ContainerInput:
    def __init__(self, name: str, cport_to: ContainerPort):
        self.name = name
        self.type_name = cport_to.port.type_name
        self.causality = cport_to.port.causality
        self.cport_list = [cport_to]
        self.vr = None

    def add_cport(self, cport_to: ContainerPort):
        if cport_to in self.cport_list: # Cannot be reached ! (Assembly prevent this to happen)
            raise FMUContainerError(f"Duplicate INPUT {cport_to} already connected to {self.name}")

        if cport_to.port.type_name != self.type_name:
            raise FMUContainerError(f"Cannot connect {self.name} of type {self.type_name} to "
                                    f"{cport_to} of type {cport_to.port.type_name}")

        if cport_to.port.causality != self.causality:
            raise FMUContainerError(f"Cannot connect {self.causality.upper()} {self.name} to "
                                    f"{cport_to.port.causality.upper()} {cport_to}")

        self.cport_list.append(cport_to)


class Link:
    CONVERSION_FUNCTION = {
        "real32/real64": "F32_F64",

        "Int8/Int16": "D8_D16",
        "Int8/UInt16": "D8_U16",
        "Int8/Int32": "D8_D32",
        "Int8/UInt32": "D8_U32",
        "Int8/Int64": "D8_D64",
        "Int8/UInt64": "D8_U64",

        "UInt8/Int16": "U8_D16",
        "UInt8/UInt16": "U8_U16",
        "UInt8/Int32": "U8_D32",
        "UInt8/UInt32": "U8_U32",
        "UInt8/Int64": "U8_D64",
        "UInt8/UInt64": "U8_U64",

        "Int16/Int32": "D16_D32",
        "Int16/UInt32": "D16_U32",
        "Int16/Int64": "D16_D64",
        "Int16/UInt64": "D16_U64",

        "UInt16/Int32": "U16_D32",
        "UInt16/UInt32": "U16_U32",
        "UInt16/Int64": "U16_D64",
        "UInt16/UInt64": "U16_U64",

        "Int32/Int64": "D32_D64",
        "Int32/UInt64": "D32_U64",

        "UInt32/Int64": "U32_D64",
        "UInt32/UInt64": "U32_U64",

        "boolean/boolean1": "B_B1",
        "boolean1/boolean": "B1_B",
    }

    def __init__(self, cport_from: ContainerPort):
        self.name = cport_from.fmu.id + "." + cport_from.port.name  # strip .fmu suffix
        self.cport_from = cport_from
        self.cport_to_list: List[ContainerPort] = []

        self.vr: Optional[int] = None
        self.vr_converted: Dict[str, Optional[int]] = {}

        if not cport_from.port.causality == "output":
            raise FMUContainerError(f"{cport_from} is  {cport_from.port.causality} instead of OUTPUT")

    def add_target(self, cport_to: ContainerPort):
        if not cport_to.port.causality == "input":
            raise FMUContainerError(f"{cport_to} is {cport_to.port.causality} instead of INPUT")

        if cport_to.port.type_name == self.cport_from.port.type_name:
            self.cport_to_list.append(cport_to)
        elif self.get_conversion(cport_to):
            self.cport_to_list.append(cport_to)
            self.vr_converted[cport_to.port.type_name] = None
        else:
            raise FMUContainerError(f"failed to connect {self.cport_from} to {cport_to} due to type.")

    def get_conversion(self, cport_to: ContainerPort) -> Optional[str]:
        try:
            conversion = f"{self.cport_from.port.type_name}/{cport_to.port.type_name}"
            return self.CONVERSION_FUNCTION[conversion]
        except KeyError:
            return None

    def nb_local(self) -> int:
        return 1+len(self.vr_converted)


class ValueReferenceTable:
    def __init__(self):
        self.vr_table:Dict[str, int] = {}
        self.masks: Dict[str, int] = {}
        self.nb_local_variable:Dict[str, int] = {}
        for i, type_name in enumerate(EmbeddedFMUPort.ALL_TYPES):
            self.vr_table[type_name] = 0
            self.masks[type_name] = i << 24
            self.nb_local_variable[type_name] = 0

    def add_vr(self, port_or_type_name: Union[ContainerPort, str], local: bool = False) -> int:
        if isinstance(port_or_type_name, ContainerPort):
            type_name = port_or_type_name.port.type_name
        else:
            type_name = port_or_type_name

        if local:
            self.nb_local_variable[type_name] += 1

        vr = self.vr_table[type_name]
        self.vr_table[type_name] += 1

        return vr | self.masks[type_name]

    def set_link_vr(self, link: Link):
        link.vr = self.add_vr(link.cport_from, local=True)
        for type_name in link.vr_converted.keys():
            link.vr_converted[type_name] = self.add_vr(type_name, local=True)

    def nb_local(self, type_name: str) -> int:
        return self.nb_local_variable[type_name]

class AutoWired:
    def __init__(self):
        self.rule_input = []
        self.rule_output = []
        self.rule_link = []
        self.nb_param = 0

    def __repr__(self):
        return (f"{self.nb_param} parameters, {len(self.rule_input) - self.nb_param} inputs,"
                f" {len(self.rule_output)} outputs, {len(self.rule_link)} links.")

    def add_input(self, from_port, to_fmu, to_port):
        self.rule_input.append([from_port, to_fmu, to_port])

    def add_parameter(self, from_port, to_fmu, to_port):
        self.rule_input.append([from_port, to_fmu, to_port])
        self.nb_param += 1

    def add_output(self, from_fmu, from_port, to_port):
        self.rule_output.append([from_fmu, from_port, to_port])

    def add_link(self, from_fmu, from_port, to_fmu, to_port):
        self.rule_link.append([from_fmu, from_port, to_fmu, to_port])


class FMUContainer:
    HEADER_XML_2 = """<?xml version="1.0" encoding="ISO-8859-1"?>
<fmiModelDescription
  fmiVersion="2.0"
  modelName="{identifier}"
  generationTool="FMUContainer-{tool_version}"
  generationDateAndTime="{timestamp}"
  guid="{guid}"
  description="FMUContainer with {embedded_fmu}"
  author="{author}"
  license="Proprietary"
  copyright="See Embedded FMU's copyrights."
  variableNamingConvention="structured">

  <CoSimulation
    modelIdentifier="{identifier}"
    canHandleVariableCommunicationStepSize="true"
    canBeInstantiatedOnlyOncePerProcess="{only_once}"
    canNotUseMemoryManagementFunctions="true"
    canGetAndSetFMUstate="false"
    canSerializeFMUstate="false"
    providesDirectionalDerivative="false"
    needsExecutionTool="{execution_tool}">
  </CoSimulation>

  <LogCategories>
    <Category name="fmucontainer"/>
  </LogCategories>

  <DefaultExperiment stepSize="{step_size}" startTime="{start_time}" stopTime="{stop_time}"/>

  <ModelVariables>
    <ScalarVariable valueReference="0" name="time" causality="independent"><Real /></ScalarVariable>
"""

    HEADER_XML_3 = """<?xml version="1.0" encoding="ISO-8859-1"?>
<fmiModelDescription
  fmiVersion="3.0"
  modelName="{identifier}"
  generationTool="FMUContainer-{tool_version}"
  generationDateAndTime="{timestamp}"
  instantiationToken="{guid}"
  description="FMUContainer with {embedded_fmu}"
  author="{author}"
  license="Proprietary"
  copyright="See Embedded FMU's copyrights."
  variableNamingConvention="structured">

  <CoSimulation
    modelIdentifier="{identifier}"
    canHandleVariableCommunicationStepSize="true"
    canBeInstantiatedOnlyOncePerProcess="{only_once}"
    canNotUseMemoryManagementFunctions="true"
    canGetAndSetFMUstate="false"
    canSerializeFMUstate="false"
    providesDirectionalDerivative="false"
    needsExecutionTool="{execution_tool}">
  </CoSimulation>

  <LogCategories>
    <Category name="fmucontainer"/>
  </LogCategories>

  <DefaultExperiment stepSize="{step_size}" startTime="{start_time}" stopTime="{stop_time}"/>

  <ModelVariables>
    <Float64 valueReference="0" name="time" causality="independent"/>
"""

    def __init__(self, identifier: str, fmu_directory: Union[str, Path], description_pathname=None, fmi_version=2):
        self.fmu_directory = Path(fmu_directory)
        self.identifier = identifier
        if not self.fmu_directory.is_dir():
            raise FMUContainerError(f"{self.fmu_directory} is not a valid directory")
        self.involved_fmu: OrderedDict[str, EmbeddedFMU] = OrderedDict()

        self.description_pathname = description_pathname
        self.fmi_version = fmi_version

        self.start_time = None
        self.stop_time = None

        # Rules
        self.inputs: Dict[str, ContainerInput] = {}
        self.outputs: Dict[str, ContainerPort] = {}
        self.links: Dict[ContainerPort, Link] = {}

        self.rules: Dict[ContainerPort, str] = {}
        self.start_values: Dict[ContainerPort, str] = {}

        self.vr_table = ValueReferenceTable()

    def get_fmu(self, fmu_filename: str) -> EmbeddedFMU:
        if fmu_filename in self.involved_fmu:
            return self.involved_fmu[fmu_filename]

        try:
            fmu = EmbeddedFMU(self.fmu_directory / fmu_filename)
            if not fmu.fmi_version == self.fmi_version:
                logger.warning(f"Try to embed FMU-{fmu.fmi_version} into container FMI-{self.fmi_version}.")
            self.involved_fmu[fmu.name] = fmu

            logger.debug(f"Adding FMU #{len(self.involved_fmu)}: {fmu}")
        except (FMUContainerError, FMUError) as e:
            raise FMUContainerError(f"Cannot load '{fmu_filename}': {e}")

        return fmu

    def mark_ruled(self, cport: ContainerPort, rule: str):
        if cport in self.rules:
            previous_rule = self.rules[cport]
            if rule not in ("OUTPUT", "LINK") and previous_rule not in ("OUTPUT", "LINK"):
                raise FMUContainerError(f"try to {rule} port {cport} which is already {previous_rule}")

        self.rules[cport] = rule

    def get_all_cports(self):
        return [ContainerPort(fmu, port_name) for fmu in self.involved_fmu.values() for port_name in fmu.ports]

    def add_input(self, container_port_name: str, to_fmu_filename: str, to_port_name: str):
        if not container_port_name:
            container_port_name = to_port_name
        cport_to = ContainerPort(self.get_fmu(to_fmu_filename), to_port_name)
        if cport_to.port.causality not in ("input", "parameter"):  # check causality
            raise FMUContainerError(f"Tried to use '{cport_to}' as INPUT of the container but FMU causality is "
                                    f"'{cport_to.port.causality}'.")

        try:
            input_port = self.inputs[container_port_name]
            input_port.add_cport(cport_to)
        except KeyError:
            self.inputs[container_port_name] = ContainerInput(container_port_name, cport_to)

        logger.debug(f"INPUT: {to_fmu_filename}:{to_port_name}")
        self.mark_ruled(cport_to, 'INPUT')

    def add_output(self, from_fmu_filename: str, from_port_name: str, container_port_name: str):
        if not container_port_name:  # empty is allowed
            container_port_name = from_port_name

        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        if cport_from.port.causality not in ("output", "local"):  # check causality
            raise FMUContainerError(f"Tried to use '{cport_from}' as OUTPUT of the container but FMU causality is "
                                    f"'{cport_from.port.causality}'.")

        if container_port_name in self.outputs:
            raise FMUContainerError(f"Duplicate OUTPUT {container_port_name} already connected to {cport_from}")

        logger.debug(f"OUTPUT: {from_fmu_filename}:{from_port_name}")
        self.mark_ruled(cport_from, 'OUTPUT')
        self.outputs[container_port_name] = cport_from

    def drop_port(self, from_fmu_filename: str, from_port_name: str):
        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        if not cport_from.port.causality == "output":  # check causality
            raise FMUContainerError(f"{cport_from}: trying to DROP {cport_from.port.causality}")

        logger.debug(f"DROP: {from_fmu_filename}:{from_port_name}")
        self.mark_ruled(cport_from, 'DROP')

    def add_link(self, from_fmu_filename: str, from_port_name: str, to_fmu_filename: str, to_port_name: str):
        cport_from = ContainerPort(self.get_fmu(from_fmu_filename), from_port_name)
        try:
            local = self.links[cport_from]
        except KeyError:
            local = Link(cport_from)

        cport_to = ContainerPort(self.get_fmu(to_fmu_filename), to_port_name)
        local.add_target(cport_to)  # Causality is check in the add() function

        logger.debug(f"LINK: {cport_from} -> {cport_to}")
        self.mark_ruled(cport_from, 'LINK')
        self.mark_ruled(cport_to, 'LINK')
        self.links[cport_from] = local

    def add_start_value(self, fmu_filename: str, port_name: str, value: str):
        cport = ContainerPort(self.get_fmu(fmu_filename), port_name)

        try:
            if cport.port.type_name in ('Real', 'Float64', 'Float32'):
                value = float(value)
            elif cport.port.type_name in ('Integer', 'Int8', 'UInt8', 'Int16', 'UInt16', 'Int32', 'UInt32', 'Int64', 'UInt64'):
                value = int(value)
            elif cport.port.type_name == 'Boolean':
                value = int(bool(value))
            else:
                value = value
        except ValueError:
            raise FMUContainerError(f"Start value is not conforming to '{cport.port.type_name}' format.")

        self.start_values[cport] = value

    def find_inputs(self, port_to_connect: EmbeddedFMUPort) -> List[ContainerPort]:
        candidates = []
        for cport in self.get_all_cports():
            if (cport.port.causality == 'input' and cport not in self.rules and cport.port.name == port_to_connect.name
                    and cport.port.type_name == port_to_connect.type_name):
                candidates.append(cport)
        return candidates

    def add_implicit_rule(self, auto_input=True, auto_output=True, auto_link=True, auto_parameter=False,
                          auto_local=False) -> AutoWired:

        auto_wired = AutoWired()
        # Auto Link outputs
        for cport in self.get_all_cports():
            if cport.port.causality == 'output':
                candidates_cport_list = self.find_inputs(cport.port)
                if auto_link and candidates_cport_list:
                    for candidate_cport in candidates_cport_list:
                        logger.info(f"AUTO LINK: {cport} -> {candidate_cport}")
                        self.add_link(cport.fmu.name, cport.port.name,
                                      candidate_cport.fmu.name, candidate_cport.port.name)
                        auto_wired.add_link(cport.fmu.name, cport.port.name,
                                            candidate_cport.fmu.name, candidate_cport.port.name)
                elif auto_output and cport not in self.rules:
                    logger.info(f"AUTO OUTPUT: Expose {cport}")
                    self.add_output(cport.fmu.name, cport.port.name, cport.port.name)
                    auto_wired.add_output(cport.fmu.name, cport.port.name, cport.port.name)
            elif cport.port.causality == 'local':
                local_portname = None
                if cport.port.name.startswith("container."):
                    local_portname = "container." + cport.fmu.id + "." + cport.port.name[10:]
                    logger.info(f"PROFILING: Expose {cport}")
                elif auto_local:
                    local_portname = cport.fmu.id + "." + cport.port.name
                    logger.info(f"AUTO LOCAL: Expose {cport}")
                if local_portname:
                    self.add_output(cport.fmu.name, cport.port.name, local_portname)
                    auto_wired.add_output(cport.fmu.name, cport.port.name, local_portname)

        if auto_input:
            # Auto link inputs
            for cport in self.get_all_cports():
                if cport not in self.rules:
                    if cport.port.causality == 'parameter' and auto_parameter:
                        parameter_name = cport.fmu.id + "." + cport.port.name
                        logger.info(f"AUTO PARAMETER: {cport} as {parameter_name}")
                        self.add_input(parameter_name, cport.fmu.name, cport.port.name)
                        auto_wired.add_parameter(parameter_name, cport.fmu.name, cport.port.name)
                    elif cport.port.causality == 'input':
                        logger.info(f"AUTO INPUT: Expose {cport}")
                        self.add_input(cport.port.name, cport.fmu.name, cport.port.name)
                        auto_wired.add_input(cport.port.name, cport.fmu.name, cport.port.name)

        logger.info(f"Auto-wiring: {auto_wired}")

        return auto_wired

    def default_step_size(self) -> float:
        freq_set = set()
        for fmu in self.involved_fmu.values():
            if fmu.step_size and fmu.capabilities["canHandleVariableCommunicationStepSize"] == "false":
                freq_set.add(int(1.0/fmu.step_size))

        common_freq = math.gcd(*freq_set)
        try:
            step_size = 1.0 / float(common_freq)
        except ZeroDivisionError:
            step_size = 0.1
            logger.warning(f"Defaulting to step_size={step_size}")

        return step_size

    def sanity_check(self, step_size: Optional[float]):
        for fmu in self.involved_fmu.values():
            if fmu.step_size and fmu.capabilities["canHandleVariableCommunicationStepSize"] == "false":
                ts_ratio = step_size / fmu.step_size
                logger.debug(f"container step_size: {step_size} = {fmu.step_size} x {ts_ratio} for {fmu.name}")
                if ts_ratio < 1.0:
                    logger.warning(f"Container step_size={step_size}s is lower than FMU '{fmu.name}' "
                                   f"step_size={fmu.step_size}s.")
                if ts_ratio != int(ts_ratio):
                    logger.warning(f"Container step_size={step_size}s should divisible by FMU '{fmu.name}' "
                                   f"step_size={fmu.step_size}s.")
            for port_name in fmu.ports:
                cport = ContainerPort(fmu, port_name)
                if cport not in self.rules:
                    if cport.port.causality == 'input':
                        logger.error(f"{cport} is not connected")
                    if cport.port.causality == 'output':
                        logger.warning(f"{cport} is not connected")

    def make_fmu(self, fmu_filename: Union[str, Path], step_size: Optional[float] = None, debug=False, mt=False,
                 profiling=False, sequential=False):
        if isinstance(fmu_filename, str):
            fmu_filename = Path(fmu_filename)

        if step_size is None:
            logger.info(f"step_size  will be deduced from the embedded FMU's")
            step_size = self.default_step_size()
        self.sanity_check(step_size)

        logger.info(f"Building FMU '{fmu_filename}', step_size={step_size}")

        base_directory = self.fmu_directory / fmu_filename.with_suffix('')
        resources_directory = self.make_fmu_skeleton(base_directory)
        with open(base_directory / "modelDescription.xml", "wt") as xml_file:
            self.make_fmu_xml(xml_file, step_size, profiling)
        with open(resources_directory / "container.txt", "wt") as txt_file:
            self.make_fmu_txt(txt_file, step_size, mt, profiling, sequential)

        self.make_fmu_package(base_directory, fmu_filename)
        if not debug:
            self.make_fmu_cleanup(base_directory)

    def make_fmu_xml(self, xml_file, step_size: float, profiling: bool):
        timestamp = datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        guid = str(uuid.uuid4())
        embedded_fmu = ", ".join([fmu_name for fmu_name in self.involved_fmu])
        try:
            author = getpass.getuser()
        except OSError:
            author = "Unspecified"

        capabilities = {}
        for capability in EmbeddedFMU.capability_list:
            capabilities[capability] = "false"
            for fmu in self.involved_fmu.values():
                if fmu.capabilities[capability] == "true":
                    capabilities[capability] = "true"

        first_fmu = next(iter(self.involved_fmu.values()))
        if self.start_time is None:
            self.start_time = first_fmu.start_time
            logger.info(f"start_time={self.start_time} (deduced from '{first_fmu.name}')")
        else:
            logger.info(f"start_time={self.start_time}")

        if self.stop_time is None:
            self.stop_time = first_fmu.stop_time
            logger.info(f"stop_time={self.stop_time} (deduced from '{first_fmu.name}')")
        else:
            logger.info(f"stop_time={self.stop_time}")

        if self.fmi_version == 2:
            xml_file.write(self.HEADER_XML_2.format(identifier=self.identifier, tool_version=tool_version,
                                                    timestamp=timestamp, guid=guid, embedded_fmu=embedded_fmu,
                                                    author=author,
                                                    only_once=capabilities['canBeInstantiatedOnlyOncePerProcess'],
                                                    execution_tool=capabilities['needsExecutionTool'],
                                                    start_time=self.start_time, stop_time=self.stop_time,
                                                    step_size=step_size))
        elif self.fmi_version == 3:
            xml_file.write(self.HEADER_XML_3.format(identifier=self.identifier, tool_version=tool_version,
                                                    timestamp=timestamp, guid=guid, embedded_fmu=embedded_fmu,
                                                    author=author,
                                                    only_once=capabilities['canBeInstantiatedOnlyOncePerProcess'],
                                                    execution_tool=capabilities['needsExecutionTool'],
                                                    start_time=self.start_time, stop_time=self.stop_time,
                                                    step_size=step_size))

        vr_time = self.vr_table.add_vr("real64", local=True)
        logger.debug(f"Time vr = {vr_time}")

        if profiling:
            for fmu in self.involved_fmu.values():
                vr = self.vr_table.add_vr("real64", local=True)
                port = EmbeddedFMUPort("real64", {"valueReference": vr,
                                        "name": f"container.{fmu.id}.rt_ratio",
                                        "description": f"RT ratio for embedded FMU '{fmu.name}'"})
                print(f"    {port.xml(vr, fmi_version=self.fmi_version)}", file=xml_file)

        index_offset = 2    # index of output ports. Start at 2 to skip "time" port

        # Local variable should be first to ensure to attribute them the lowest VR.
        for link in self.links.values():
            self.vr_table.set_link_vr(link)
            port_local_def = link.cport_from.port.xml(link.vr, name=link.name, causality='local',
                                                      fmi_version=self.fmi_version)
            if port_local_def:
                print(f"    {port_local_def}", file=xml_file)
                index_offset += 1

        for input_port_name, input_port in self.inputs.items():
            input_port.vr = self.vr_table.add_vr(input_port.type_name)
            # Get Start and XML from first connected input
            start = self.start_values.get(input_port.cport_list[0], None)
            port_input_def = input_port.cport_list[0].port.xml(input_port.vr, name=input_port_name,
                                                               start=start, fmi_version=self.fmi_version)
            if port_input_def:
                print(f"    {port_input_def}", file=xml_file)
                index_offset += 1

        for output_port_name, output_port in self.outputs.items():
            output_port.vr = self.vr_table.add_vr(output_port)
            port_output_def = output_port.port.xml(output_port.vr, name=output_port_name,
                                                   fmi_version=self.fmi_version)
            if port_output_def:
                print(f"    {port_output_def}", file=xml_file)

        if self.fmi_version == 2:
            self.make_fmu_xml_epilog_2(xml_file, index_offset)
        elif self.fmi_version == 3:
            self.make_fmu_xml_epilog_3(xml_file)

    def make_fmu_xml_epilog_2(self, xml_file, index_offset):
        xml_file.write("  </ModelVariables>\n"
                       "\n"
                       "  <ModelStructure>\n"
                       "    <Outputs>\n")

        index = index_offset
        for output in self.outputs.values():
            if output.port.type_name in EmbeddedFMUPort.CONTAINER_TO_FMI[2]:
                print(f'      <Unknown index="{index}"/>', file=xml_file)
                index += 1
        xml_file.write("    </Outputs>\n"
                       "    <InitialUnknowns>\n")
        index = index_offset
        for output in self.outputs.values():
            if output.port.type_name in EmbeddedFMUPort.CONTAINER_TO_FMI[2]:
                print(f'      <Unknown index="{index}"/>', file=xml_file)
                index += 1
        xml_file.write("    </InitialUnknowns>\n"
                       "  </ModelStructure>\n"
                       "\n"
                       "</fmiModelDescription>")

    def make_fmu_xml_epilog_3(self, xml_file):
        xml_file.write("  </ModelVariables>\n"
                       "\n"
                       "  <ModelStructure>\n")
        for output in self.outputs.values():
            if output.port.type_name in EmbeddedFMUPort.CONTAINER_TO_FMI[3]:
                print(f'      <Output valueReference="{output.vr}"/>', file=xml_file)
        for output in self.outputs.values():
            if output.port.type_name in EmbeddedFMUPort.CONTAINER_TO_FMI[3]:
                print(f'      <InitialUnknown valueReference="{output.vr}"/>', file=xml_file)
        xml_file.write("  </ModelStructure>\n"
                       "\n"
                       "</fmiModelDescription>")

    def make_fmu_txt(self, txt_file, step_size: float, mt: bool, profiling: bool, sequential: bool):
        print("# Container flags <MT> <Profiling> <Sequential>", file=txt_file)
        flags = [ str(int(flag == True)) for flag in (mt, profiling, sequential)]
        print(" ".join(flags), file=txt_file)

        print(f"# Internal time step in seconds", file=txt_file)
        print(f"{step_size}", file=txt_file)
        print(f"# NB of embedded FMU's", file=txt_file)
        print(f"{len(self.involved_fmu)}", file=txt_file)
        fmu_rank: Dict[str, int] = {}
        for i, fmu in enumerate(self.involved_fmu.values()):
            print(f"{fmu.name} {fmu.fmi_version}", file=txt_file)
            print(f"{fmu.model_identifier}", file=txt_file)
            print(f"{fmu.guid}", file=txt_file)
            fmu_rank[fmu.name] = i

        # Prepare data structure
        inputs_per_type: Dict[str, List[ContainerInput]] = {}       # Container's INPUT
        outputs_per_type: Dict[str, List[ContainerPort]] = {}       # Container's OUTPUT

        inputs_fmu_per_type: Dict[str, Dict[str, Dict[ContainerPort, int]]] = {}      # [type][fmu]
        start_values_fmu_per_type = {}
        outputs_fmu_per_type = {}
        local_per_type: Dict[str, List[int]] = {}
        links_per_fmu: Dict[str, List[Link]] = {}

        for type_name in EmbeddedFMUPort.ALL_TYPES:
            inputs_per_type[type_name] = []
            outputs_per_type[type_name] = []
            local_per_type[type_name] = []

            inputs_fmu_per_type[type_name] = {}
            start_values_fmu_per_type[type_name] = {}
            outputs_fmu_per_type[type_name] = {}

            for fmu in self.involved_fmu.values():
                inputs_fmu_per_type[type_name][fmu.name] = {}
                start_values_fmu_per_type[type_name][fmu.name] = {}
                outputs_fmu_per_type[type_name][fmu.name] = {}

        # Fill data structure
        # Inputs
        for input_port_name, input_port in self.inputs.items():
            inputs_per_type[input_port.type_name].append(input_port)
        for input_port, value in self.start_values.items():
            start_values_fmu_per_type[input_port.port.type_name][input_port.fmu.name][input_port] = value
        # Outputs
        for output_port_name, output_port in self.outputs.items():
            outputs_per_type[output_port.port.type_name].append(output_port)
        # Links
        for link in self.links.values():
            local_per_type[link.cport_from.port.type_name].append(link.vr)
            outputs_fmu_per_type[link.cport_from.port.type_name][link.cport_from.fmu.name][link.cport_from] = link.vr
            for cport_to in link.cport_to_list:
                if cport_to.port.type_name == link.cport_from.port.type_name:
                    inputs_fmu_per_type[cport_to.port.type_name][cport_to.fmu.name][cport_to] = link.vr
                else:
                    local_per_type[cport_to.port.type_name].append(link.vr_converted[cport_to.port.type_name])
                    links_per_fmu.setdefault(link.cport_from.fmu.name, []).append(link)
                    inputs_fmu_per_type[cport_to.port.type_name][cport_to.fmu.name][cport_to] = link.vr_converted[cport_to.port.type_name]


        print(f"# NB local variables:", ", ".join(EmbeddedFMUPort.ALL_TYPES), file=txt_file)
        nb_local = [f"{self.vr_table.nb_local(type_name)}" for type_name in EmbeddedFMUPort.ALL_TYPES]
        print(" ".join(nb_local), file=txt_file, end='')
        print("", file=txt_file)

        print("# CONTAINER I/O: <VR> <NB> <FMU_INDEX> <FMU_VR> [<FMU_INDEX> <FMU_VR>]", file=txt_file)
        for type_name in EmbeddedFMUPort.ALL_TYPES:
            print(f"# {type_name}" , file=txt_file)
            nb_local = (len(inputs_per_type[type_name]) +
                        len(outputs_per_type[type_name]) +
                        self.vr_table.nb_local(type_name))
            nb_input_link = 0
            for input_port in inputs_per_type[type_name]:
                nb_input_link += len(input_port.cport_list) - 1
            print(f"{nb_local} {nb_local + nb_input_link}", file=txt_file)
            if type_name == "real64":
                print(f"0 1 -1 0", file=txt_file)  # Time slot
                if profiling:
                    for profiling_port, _ in enumerate(self.involved_fmu.values()):
                        print(f"{profiling_port + 1} 1 -2 {profiling_port + 1}", file=txt_file)

            for input_port in inputs_per_type[type_name]:
                cport_string = [f"{fmu_rank[cport.fmu.name]} {cport.port.vr}" for cport in input_port.cport_list]
                print(f"{input_port.vr} {len(input_port.cport_list)}", " ".join(cport_string), file=txt_file)
            for output_port in outputs_per_type[type_name]:
                print(f"{output_port.vr} 1 {fmu_rank[output_port.fmu.name]} {output_port.port.vr}", file=txt_file)
            for local_vr in local_per_type[type_name]:
                print(f"{local_vr} 1 -1 {local_vr & 0xFFFFFF}", file=txt_file)

        # LINKS
        for fmu in self.involved_fmu.values():
            for type_name in EmbeddedFMUPort.ALL_TYPES:
                print(f"# Inputs of {fmu.name} - {type_name}: <VR> <FMU_VR>", file=txt_file)
                print(len(inputs_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for input_port, vr in inputs_fmu_per_type[type_name][fmu.name].items():
                    print(f"{vr} {input_port.port.vr}", file=txt_file)

            for type_name in EmbeddedFMUPort.ALL_TYPES:
                print(f"# Start values of {fmu.name} - {type_name}: <FMU_VR> <RESET> <VALUE>", file=txt_file)
                print(len(start_values_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for input_port, value in start_values_fmu_per_type[type_name][fmu.name].items():
                    reset = 1 if input_port.port.causality == "input" else 0
                    print(f"{input_port.port.vr} {reset} {value}", file=txt_file)

            for type_name in EmbeddedFMUPort.ALL_TYPES:
                print(f"# Outputs of {fmu.name} - {type_name}: <VR> <FMU_VR>", file=txt_file)
                print(len(outputs_fmu_per_type[type_name][fmu.name]), file=txt_file)
                for output_port, vr in outputs_fmu_per_type[type_name][fmu.name].items():
                    print(f"{vr} {output_port.port.vr}", file=txt_file)

            print(f"# Conversion table of {fmu.name}: <VR_FROM> <VR_TO> <CONVERSION>", file=txt_file)
            try:
                nb = 0
                for link in links_per_fmu[fmu.name]:
                    nb += len(link.vr_converted)
                print(f"{nb}", file=txt_file)
                for link in links_per_fmu[fmu.name]:
                    for cport_to in link.cport_to_list:
                        conversion =  link.get_conversion(cport_to)
                        if conversion:
                            print(f"{link.vr} {link.vr_converted[cport_to.port.type_name]} {conversion}",
                                  file=txt_file)
            except KeyError:
                print("0", file=txt_file)

    @staticmethod
    def long_path(path: Union[str, Path]) -> str:
        # https://stackoverflow.com/questions/14075465/copy-a-file-with-a-too-long-path-to-another-directory-in-python
        if os.name == 'nt':
            return "\\\\?\\" + os.path.abspath(str(path))
        else:
            return path

    @staticmethod
    def copyfile(origin, destination):
        logger.debug(f"Copying {origin} in {destination}")
        shutil.copy(origin, destination)

    def get_bindir_and_suffixe(self) -> Tuple[str, str, str]:
        suffixes = {
            "Windows": "dll",
            "Linux": "so",
            "Darwin": "dylib"
        }

        origin_bindirs = {
            "Windows": "win64",
            "Linux": "linux64",
            "Darwin": "darwin64"
        }

        if self.fmi_version == 3:
            target_bindirs = {
                "Windows": "x86_64-windows",
                "Linux": "x86_64-linux",
                "Darwin": "aarch64-darwin"
            }
        else:
            target_bindirs = origin_bindirs

        os_name = platform.system()
        try:
            return origin_bindirs[os_name], suffixes[os_name], target_bindirs[os_name]
        except KeyError:
            raise FMUContainerError(f"OS '{os_name}' is not supported.")

    def make_fmu_skeleton(self, base_directory: Path) -> Path:
        logger.debug(f"Initialize directory '{base_directory}'")

        origin = Path(__file__).parent / "resources"
        resources_directory = base_directory / "resources"
        documentation_directory = base_directory / "documentation"
        binaries_directory = base_directory / "binaries"

        base_directory.mkdir(exist_ok=True)
        resources_directory.mkdir(exist_ok=True)
        binaries_directory.mkdir(exist_ok=True)
        documentation_directory.mkdir(exist_ok=True)

        if self.description_pathname:
            self.copyfile(self.description_pathname, documentation_directory)

        self.copyfile(origin / "model.png", base_directory)

        origin_bindir, suffixe, target_bindir = self.get_bindir_and_suffixe()

        library_filename = origin / origin_bindir / f"container.{suffixe}"
        if not library_filename.is_file():
            raise FMUContainerError(f"File {library_filename} not found")
        binary_directory = binaries_directory / target_bindir
        binary_directory.mkdir(exist_ok=True)
        self.copyfile(library_filename, binary_directory / f"{self.identifier}.{suffixe}")

        for i, fmu in enumerate(self.involved_fmu.values()):
            shutil.copytree(self.long_path(fmu.fmu.tmp_directory),
                            self.long_path(resources_directory / f"{i:02x}"), dirs_exist_ok=True)

        return resources_directory

    def make_fmu_package(self, base_directory: Path, fmu_filename: Path):
        logger.debug(f"Zipping directory '{base_directory}' => '{fmu_filename}'")
        zip_directory = self.long_path(str(base_directory.absolute()))
        offset = len(zip_directory) + 1
        with zipfile.ZipFile(self.fmu_directory / fmu_filename, "w", zipfile.ZIP_DEFLATED) as zip_file:
            def add_file(directory: Path):
                for entry in directory.iterdir():
                    if entry.is_dir():
                        add_file(directory / entry)
                    elif entry.is_file:
                        zip_file.write(str(entry), str(entry)[offset:])

            add_file(Path(zip_directory))
        logger.info(f"'{fmu_filename}' is available.")

    def make_fmu_cleanup(self, base_directory: Path):
        logger.debug(f"Delete directory '{base_directory}'")
        shutil.rmtree(self.long_path(base_directory))



import argparse
import sys

from .utils import setup_logger, make_wide
from ..operations import (OperationSummary, OperationError, OperationRemoveRegexp,
                          OperationRemoveSources, OperationTrimUntil, OperationKeepOnlyRegexp, OperationMergeTopLevel,
                          OperationStripTopLevel, OperationRenameFromCSV, OperationSaveNamesToCSV, FMU, FMUError)
from ..remoting import  (OperationAddFrontendWin32, OperationAddFrontendWin64, OperationAddRemotingWin32,
                         OperationAddRemotingWin64)
from ..checker import checker_list
from ..version import __version__ as version
from ..help import Help


def fmutool():
    logger = setup_logger()

    logger.info(f"FMU Manipulation Toolbox version {version}")
    help_message = Help()

    parser = argparse.ArgumentParser(prog='fmutool',
                                     description="Analyse and Manipulate a FMU by modifying its 'modelDescription.xml'",
                                     formatter_class=make_wide(argparse.HelpFormatter),
                                     add_help=False,
                                     epilog="see: https://github.com/grouperenault/fmu_manipulation_toolbox/blob/main/README.md")

    def add_option(option, *args, **kwargs):
        parser.add_argument(option, *args, help=help_message.usage(option), **kwargs)

    add_option('-h', '-help', action="help")

    # I/O
    add_option('-input', action='store', dest='fmu_input', default=None, required=True, metavar='path/to/module.fmu')
    add_option('-output', action='store', dest='fmu_output', default=None, metavar='path/to/module-modified.fmu')

    # Port name manipulation
    add_option('-remove-toplevel', action='append_const', dest='operations_list', const=OperationStripTopLevel())
    add_option('-merge-toplevel', action='append_const', dest='operations_list', const=OperationMergeTopLevel())
    add_option('-trim-until', action='append', dest='operations_list', type=OperationTrimUntil, metavar='prefix')
    add_option('-remove-regexp', action='append', dest='operations_list', type=OperationRemoveRegexp,
               metavar='regular-expression')
    add_option('-keep-only-regexp', action='append', dest='operations_list', type=OperationKeepOnlyRegexp,
               metavar='regular-expression')
    add_option('-remove-all', action='append_const', dest='operations_list', const=OperationRemoveRegexp('.*'))

    # Batch Rename
    add_option('-dump-csv', action='append', dest='operations_list', type=OperationSaveNamesToCSV,
               metavar='path/to/list.csv')
    add_option('-rename-from-csv', action='append', dest='operations_list', type=OperationRenameFromCSV,
               metavar='path/to/translation.csv')

    # Remoting
    add_option('-add-remoting-win32', action='append_const', dest='operations_list', const=OperationAddRemotingWin32())
    add_option('-add-remoting-win64', action='append_const', dest='operations_list', const=OperationAddRemotingWin64())
    add_option('-add-frontend-win32', action='append_const', dest='operations_list', const=OperationAddFrontendWin32())
    add_option('-add-frontend-win64', action='append_const', dest='operations_list', const=OperationAddFrontendWin64())

    # Extraction / Removal
    add_option('-extract-descriptor', action='store', dest='extract_description',
               metavar='path/to/saved-modelDescriptor.xml')
    add_option('-remove-sources', action='append_const', dest='operations_list',
               const=OperationRemoveSources())
    # Filter
    add_option('-only-parameters', action='append_const', dest='apply_on', const='parameter')
    add_option('-only-inputs', action='append_const', dest='apply_on', const='input')
    add_option('-only-outputs', action='append_const', dest='apply_on', const='output')
    # Checker
    add_option('-summary', action='append_const', dest='operations_list', const=OperationSummary())
    add_option('-check', action='append_const', dest='operations_list', const=[checker() for checker in checker_list])

    cli_options = parser.parse_args()
    # handle the "no operation" use case
    if not cli_options.operations_list:
        cli_options.operations_list = []

    if cli_options.fmu_input == cli_options.fmu_output:
        logger.fatal(f"'-input' and '-output' should point to different files.")
        sys.exit(-3)

    logger.info(f"READING Input='{cli_options.fmu_input}'")
    try:
        fmu = FMU(cli_options.fmu_input)
    except FMUError as reason:
        logger.fatal(f"{reason}")
        sys.exit(-4)

    if cli_options.apply_on:
        logger.info("Applying operation for :")
        for causality in cli_options.apply_on:
            logger.info(f"     - causality = {causality}")

    for operation in cli_options.operations_list:
        logger.info(f"     => {operation}")
        try:
            fmu.apply_operation(operation, cli_options.apply_on)
        except OperationError as reason:
            logger.fatal(f"{reason}")
            sys.exit(-6)

    if cli_options.extract_description:
        logger.info(f"WRITING ModelDescriptor='{cli_options.extract_description}'")
        fmu.save_descriptor(cli_options.extract_description)

    if cli_options.fmu_output:
        logger.info(f"WRITING Output='{cli_options.fmu_output}'")
        try:
            fmu.repack(cli_options.fmu_output)
        except FMUError as reason:
            logger.fatal(f"FATAL ERROR: {reason}")
            sys.exit(-5)
    else:
        logger.info(f"INFO    Modified FMU is not saved. If necessary use '-output' option.")

if __name__ == "__main__":
    fmutool()



import logging
import zipfile
import json
import xml.parsers.expat

from typing import *
from pathlib import Path

logger = logging.getLogger("fmu_manipulation_toolbox")


class FMUSplitterError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return str(self.message)


class FMUSplitterPort:
    def __init__(self, fmu_filename: str, port_name):
        self.fmu_filename = fmu_filename
        self.port_name = port_name

    def __str__(self):
        return f"{self.fmu_filename}/{self.port_name}"


class FMUSplitterLink:
    def __init__(self):
        self.from_port: Optional[FMUSplitterPort] = None
        self.to_port: List[FMUSplitterPort] = []

    def __str__(self):
        to_str = ", ".join([f"{port}" for port in self.to_port])
        return f"{self.from_port} -> {to_str}"


class FMUSplitter:
    def __init__(self, fmu_filename: str):
        self.fmu_filename = Path(fmu_filename)
        self.directory = self.fmu_filename.with_suffix(".dir")
        self.zip = None # to handle error
        self.zip = zipfile.ZipFile(self.fmu_filename)
        self.filenames_list = self.zip.namelist()
        self.dir_set = self.get_dir_set()

        if "resources/container.txt" not in self.filenames_list:
                raise FMUSplitterError(f"FMU file {self.fmu_filename} is not an FMU Container.")

        self.directory.mkdir(exist_ok=True)
        logger.info(f"Preparing to split '{self.fmu_filename}' into '{self.directory}'")

    def get_dir_set(self) -> Set[str]:
        dir_set = set()
        for filename in self.filenames_list:
            parent = "/".join(filename.split("/")[:-1])
            dir_set.add(parent+"/")
        return dir_set

    def __del__(self):
        if self.zip is not None:
            logger.debug("Closing zip file")
            self.zip.close()

    def split_fmu(self):
        logger.info(f"Splitting...")
        config = self._split_fmu(fmu_filename=str(self.fmu_filename), relative_path="")
        config_filename = self.directory / self.fmu_filename.with_suffix(".json")
        with open(config_filename, "w") as file:
            json.dump(config, file, indent=2)
        logger.info(f"Container definition saved to '{config_filename}'")

    def _split_fmu(self, fmu_filename: str, relative_path: str) -> Dict[str, Any]:
        txt_filename = f"{relative_path}resources/container.txt"

        if txt_filename in self.filenames_list:
            description = FMUSplitterDescription(self.zip)
            config = description.parse_txt_file(txt_filename)
            config["name"] = fmu_filename
            for i, fmu_filename in enumerate(config["candidate_fmu"]):
                directory = f"{relative_path}resources/{i:02x}/"
                if directory not in self.dir_set:
                    directory = f"{relative_path}resources/{fmu_filename}/"
                    if directory not in self.dir_set:
                        raise FMUSplitterError(f"{directory} not found in FMU")
                sub_config = self._split_fmu(fmu_filename=fmu_filename, relative_path=directory)
                if isinstance(sub_config, str):
                    key = "fmu"
                else:
                    key = "container"
                try:
                    config[key].append(sub_config)
                except KeyError:
                    config[key] = [sub_config]
            config.pop("candidate_fmu")
        else:
            config = fmu_filename
            self.extract_fmu(relative_path, str(self.directory / fmu_filename))

        return config

    def extract_fmu(self, directory: str, filename: str):
        logger.debug(f"Extracting {directory} to {filename}")
        with zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED) as fmu_file:
            for file in self.zip.namelist():
                if file.startswith(directory):
                    data = self.zip.read(file)
                    fmu_file.writestr(file[len(directory):], data)
        logger.info(f"FMU Extraction of '{filename}'")


class FMUSplitterDescription:
    def __init__(self, zip):
        self.zip = zip
        self.links: Dict[str, Dict[int, FMUSplitterLink]] = {
            "Real": {},
            "Integer": {},
            "Boolean": {},
            "String": {}
        }
        self.vr_to_name: Dict[str, Dict[str, Dict[int, Dict[str, str]]]] = {} # name, fmi_type, vr <-> {name, causality}
        self.config: Dict[str, Any] = {
            "auto_input": False,
            "auto_output": False,
            "auto_parameter": False,
            "auto_local": False,
            "auto_link": False,
        }
        self.fmu_filename_list = []

        # used for modelDescription.xml parsing
        self.current_fmu_filename = None
        self.current_vr = None
        self.current_name = None
        self.current_causality = None

    @staticmethod
    def get_line(file):
        for line in file:
            line = line.decode('utf-8').strip()
            if line and not line.startswith("#"):
                return line
        raise StopIteration

    def start_element(self, name, attrs):
        if name == "ScalarVariable":
            self.current_name = attrs["name"]
            self.current_vr = int(attrs["valueReference"])
            self.current_causality = attrs["causality"]
        elif name in self.vr_to_name[self.current_fmu_filename]:
            self.vr_to_name[self.current_fmu_filename][name][self.current_vr] = {"name": self.current_name,
                                                                                 "causality": self.current_causality}
        else:
            self.current_vr = None
            self.current_name = None
            self.current_causality = None

    def parse_model_description(self, directory: str, fmu_filename: str):

        if directory == ".":
            filename = "modelDescription.xml"
        else:
            filename = f"{directory}/modelDescription.xml"

        self.vr_to_name[fmu_filename] = {
            "Real": {},
            "Integer": {},
            "Boolean": {},
            "String": {}
        }
        parser = xml.parsers.expat.ParserCreate()
        self.current_fmu_filename = fmu_filename
        self.current_vr = None
        self.current_name = None
        self.current_causality = None
        parser.StartElementHandler = self.start_element
        with (self.zip.open(filename) as file):
            logger.debug(f"Parsing '{filename}' ({fmu_filename})")
            parser.ParseFile(file)

    def parse_txt_file_header(self, file, txt_filename):
        flags = self.get_line(file).split(" ")
        if len(flags) == 1:
            self.config["mt"] = flags[0] == "1"
            self.config["profiling"] = self.get_line(file) == "1"
            self.config["sequential"] = False
        elif len(flags) == 3:
            self.config["mt"] = flags[0] == "1"
            self.config["profiling"] = flags[1] == "1"
            self.config["sequential"] = flags[2] == "1"
        else:
            raise FMUSplitterError(f"Cannot interpret flags '{flags}'")

        self.config["step_size"] = float(self.get_line(file))
        nb_fmu = int(self.get_line(file))
        logger.debug(f"Number of FMUs: {nb_fmu}")
        self.config["candidate_fmu"] = []

        for i in range(nb_fmu):
            # format is
            #    filename.fmu
            # or
            #    filename.fmu fmi_version
            fmu_filename = self.get_line(file)
            if ' ' in fmu_filename:
                fmu_filename = fmu_filename.split(' ')[0]
                # fmi version is not needed for further operations

            base_directory = "/".join(txt_filename.split("/")[0:-1])

            directory = f"{base_directory}/{fmu_filename}"
            if f"{directory}/modelDescription.xml" not in self.zip.namelist():
                directory = f"{base_directory}/{i:02x}"
                if f"{directory}/modelDescription.xml" not in self.zip.namelist():
                    print(self.zip.namelist())
                    raise FMUSplitterError(f"Cannot find in FMU directory in {directory}.")

            self.parse_model_description(directory, fmu_filename)
            self.config["candidate_fmu"].append(fmu_filename)
            _library = self.get_line(file)
            _uuid = self.get_line(file)
        _nb_local_variables = self.get_line(file)

    def add_port(self, fmi_type: str, fmu_id: int, fmu_vr: int, container_vr: int):
        if fmu_id >= 0:
            fmu_filename = self.config["candidate_fmu"][fmu_id]
            causality = self.vr_to_name[fmu_filename][fmi_type][fmu_vr]["causality"]
            fmu_port_name = self.vr_to_name[fmu_filename][fmi_type][fmu_vr]["name"]
            container_port = self.vr_to_name["."][fmi_type][container_vr]["name"]
            if not causality == "output":
                causality = "input"
                definition = [container_port, fmu_filename, fmu_port_name]
            else:
                definition = [fmu_filename, fmu_port_name, container_port]

            try:
                self.config[causality].append(definition)
            except KeyError:
                self.config[causality] = [definition]

            logger.debug(f"Adding container port {causality}: {definition}")

    def parse_txt_file_ports(self, file):
        for fmi_type in ("Real", "Integer", "Boolean", "String"):
            nb_port_variables = self.get_line(file).split(" ")[0]
            for i in range(int(nb_port_variables)):
                tokens = self.get_line(file).split(" ")
                if len(tokens) > 3:
                    container_vr = int(tokens[0])
                    for j in range(int(tokens[1])):
                        fmu_id = int(tokens[2 + 2 * j])
                        fmu_vr = int(tokens[2 + 2 * j + 1])
                        self.add_port(fmi_type, fmu_id, fmu_vr, container_vr)
                else:  # For FMUContainer <= 1.8.4
                    container_vr = int(tokens[0])
                    fmu_id = int(tokens[1])
                    fmu_vr = int(tokens[2])
                    self.add_port(fmi_type, fmu_id, fmu_vr, container_vr)

    def parse_txt_file(self, txt_filename: str):
        self.parse_model_description(str(Path(txt_filename).parent.parent), ".")
        logger.debug(f"Parsing container file '{txt_filename}'")

        with (self.zip.open(txt_filename) as file):
            self.parse_txt_file_header(file, txt_filename)
            self.parse_txt_file_ports(file)

            for fmu_filename in self.config["candidate_fmu"]:
                # Inputs per FMUs
                for fmi_type in ("Real", "Integer", "Boolean", "String"):
                    nb_input = int(self.get_line(file))
                    for i in range(nb_input):
                        local, vr = self.get_line(file).split(" ")
                        try:
                            link = self.links[fmi_type][local]
                        except KeyError:
                            link = FMUSplitterLink()
                            self.links[fmi_type][local] = link
                        link.to_port.append(FMUSplitterPort(fmu_filename,
                                                            self.vr_to_name[fmu_filename][fmi_type][int(vr)]["name"]))

                for fmi_type in ("Real", "Integer", "Boolean", "String"):
                    nb_start = int(self.get_line(file))
                    for i in range(nb_start):
                        tokens = self.get_line(file).split(" ")
                        vr = int(tokens[0])
                        value = tokens[-1]
                        start_definition = [fmu_filename, self.vr_to_name[fmu_filename][fmi_type][vr]["name"],
                                            value]
                        try:
                            self.config["start"].append(start_definition)
                        except KeyError:
                            self.config["start"] = [start_definition]

                # Output per FMUs
                for fmi_type in ("Real", "Integer", "Boolean", "String"):
                    nb_output = int(self.get_line(file))

                    for i in range(nb_output):
                        local, vr = self.get_line(file).split(" ")
                        try:
                            link = self.links[fmi_type][local]
                        except KeyError:
                            link = FMUSplitterLink()
                            self.links[fmi_type][local] = link
                        link.from_port = FMUSplitterPort(fmu_filename,
                                                         self.vr_to_name[fmu_filename][fmi_type][int(vr)]["name"])

            logger.debug("End of parsing.")

        for fmi_type, links in self.links.items():
            for link in links.values():
                for to_port in link.to_port:
                    definition = [ link.from_port.fmu_filename, link.from_port.port_name,
                                   to_port.fmu_filename, to_port.port_name]
                    try:
                        self.config["link"].append(definition)
                    except KeyError:
                        self.config["link"] = [definition]

        return self.config



import os
import re
from setuptools import setup

from fmu_manipulation_toolbox.version import __author__ as author, __version__ as default_version

try:
    version = os.environ["GITHUB_REF_NAME"]
except Exception as e:
    print(f"Cannot get repository status: {e}. Defaulting to {default_version}")
    version = default_version

if not re.match(r"[A-Za-z]?\d+(\.\d)+", version):
    print(f"WARNING: Version {version} does not match standard. The publication will fail !")
    version = default_version

# Create __version__.py
try:
    with open("fmu_manipulation_toolbox/__version__.py", "wt") as file:
        print(f"'{version}'", file=file)
except Exception as e:
    print(f"Cannot create __version__.py: {e}")

setup(
    name="fmu_manipulation_toolbox",
    version=version,
    packages=["fmu_manipulation_toolbox", "fmu_manipulation_toolbox.cli"],
    package_data={"fmu_manipulation_toolbox": [
        "resources/win32/client_sm.dll",
        "resources/win32/server_sm.exe",
        "resources/win64/client_sm.dll",
        "resources/win64/server_sm.exe",
        "resources/win64/container.dll",
        "resources/linux64/client_sm.so",
        "resources/linux64/server_sm",
        "resources/linux64/container.so",
        "resources/linux32/client_sm.so",
        "resources/linux32/server_sm",
        "resources/darwin64/container.dylib",
        "resources/license.txt",
        "resources/*.png",
        "resources/fmi-2.0/*.xsd",
        "resources/fmi-3.0/*.xsd",
    ]},
    entry_points={"console_scripts": ["fmutool = fmu_manipulation_toolbox.cli.fmutool:fmutool",
                                      "fmucontainer = fmu_manipulation_toolbox.cli.fmucontainer:fmucontainer",
                                      "fmusplit = fmu_manipulation_toolbox.cli.fmusplit:fmusplit"
                                      ],
                    "gui_scripts": ["fmutool-gui = fmu_manipulation_toolbox.gui:main"]
                  },
    author=author,
    url="https://github.com/grouperenault/fmu_manipulation_toolbox/",
    description="FMU Manipulation Toolbox is a python application for modifying Functional Mock-up Units "
                "(FMUs) without recompilation or bundling them into FMU Containers",
    long_description="""FMU Manipulation Toolbox is a python application for modifying Functional Mock-up Units (FMUs) 
without recompilation. It mainly modifies the `modelDescription.xml` file. It is highly customizable.

Manipulating the `modelDescription.xml` can be a dangerous thing! Communicating with the FMU-developer and adapting
the way the FMU is generated, is the preferable when possible.

FMU Manipulation Toolbox also allows to group FMU's inside FMU Containers.
    """,
    install_requires=[
        "PySide6 >= 6.8.0",
        "xmlschema >= 3.3.1",
        "elementpath >= 4.4.0",
        "colorama >= 0.4.6",
    ],
    license="BSD-2-Clause",
    python_requires=">=3.8",
)

os.remove("fmu_manipulation_toolbox/__version__.py")



import logging
import shutil

from pathlib import Path
from .operations import OperationAbstract, OperationError

logger = logging.getLogger("fmu_manipulation_toolbox")

class OperationAddRemotingWinAbstract(OperationAbstract):
    bitness_from = None
    bitness_to = None

    def __repr__(self):
        return f"Add '{self.bitness_to}' remoting on '{self.bitness_from}' FMU"

    def __init__(self):
        self.vr = {
            "Real": [],
            "Integer": [],
            "Boolean": []
        }
        self.nb_input = 0;
        self.nb_output = 0;

    def fmi_attrs(self, attrs):
        if not attrs["fmiVersion"] == "2.0":
            raise OperationError(f"Adding remoting is only available for FMI-2.0")

    def cosimulation_attrs(self, attrs):
        fmu_bin = {
            "win32": Path(self.fmu.tmp_directory) / "binaries" / "win32",
            "win64": Path(self.fmu.tmp_directory) / "binaries" / "win64",
        }

        if not fmu_bin[self.bitness_from].is_dir():
            raise OperationError(f"{self.bitness_from} interface does not exist")

        if fmu_bin[self.bitness_to].is_dir():
            logger.info(f"{self.bitness_to} already exists. Add front-end.")
            shutil.move(fmu_bin[self.bitness_to] / Path(attrs['modelIdentifier']).with_suffix(".dll"),
                        fmu_bin[self.bitness_to] / Path(attrs['modelIdentifier']).with_suffix("-remoted.dll"))
        else:
            fmu_bin[self.bitness_to].mkdir()

        to_path = Path(__file__).parent / "resources" / self.bitness_to
        try:
            shutil.copyfile(to_path / "client_sm.dll",
                            fmu_bin[self.bitness_to] / Path(attrs['modelIdentifier']).with_suffix(".dll"))
        except FileNotFoundError as e:
            logger.critical(f"Cannot add remoting client: {e}")

        from_path = Path(__file__).parent / "resources" / self.bitness_from
        try:
            shutil.copyfile(from_path / "server_sm.exe",
                            fmu_bin[self.bitness_from] / "server_sm.exe")
        except FileNotFoundError as e:
            logger.critical(f"Cannot add remoting server: {e}")

        shutil.copyfile(Path(__file__).parent / "resources" / "license.txt",
                        fmu_bin[self.bitness_to] / "license.txt")

    def port_attrs(self, fmu_port) -> int:
        try:
            self.vr[fmu_port.fmi_type].append(fmu_port["valueReference"])
            if fmu_port["causality"] in ("input", "parameter"):
                self.nb_input += 1
            else:
                self.nb_output += 1
        except KeyError:
            logger.error(f"Type '{fmu_port.fmi_type}' is not supported by remoting.")

        return 0

    def closure(self):
        target_dir = Path(self.fmu.tmp_directory) / "resources"
        if not target_dir.is_dir():
            target_dir.mkdir()

        logger.info(f"Remoting nb input port: {self.nb_input}")
        logger.info(f"Remoting nb output port: {self.nb_output}")
        with open(target_dir/ "remoting_table.txt", "wt") as file:
            for fmi_type in ('Real', 'Integer', 'Boolean'):
                print(len(self.vr[fmi_type]), file=file)
            for fmi_type in ('Real', 'Integer', 'Boolean'):
                for vr in sorted(self.vr[fmi_type]):
                    print(vr, file=file)

class OperationAddRemotingWin64(OperationAddRemotingWinAbstract):
    bitness_from = "win32"
    bitness_to = "win64"


class OperationAddFrontendWin32(OperationAddRemotingWinAbstract):
    bitness_from = "win32"
    bitness_to = "win32"


class OperationAddFrontendWin64(OperationAddRemotingWinAbstract):
    bitness_from = "win64"
    bitness_to = "win64"


class OperationAddRemotingWin32(OperationAddRemotingWinAbstract):
    bitness_from = "win64"
    bitness_to = "win32"
